@@ -12,6 +12,7 @@ from eflow.data_pipeline_segments import DataEncoder
 from sklearn.decomposition import PCA
 from scipy.cluster.hierarchy import linkage, dendrogram,set_link_color_palette
 from sklearn.preprocessing import StandardScaler
+from sklearn.metrics import silhouette_score
 from sklearn.cluster import AgglomerativeClustering
 from sklearn.neighbors import kneighbors_graph
 
@@ -20,6 +21,7 @@ from pyclustering.cluster.kmeans import kmeans
 from pyclustering.cluster.kmedians import kmedians
 from pyclustering.cluster.kmedoids import kmedoids
 from pyclustering.cluster.ema import ema
+from pyclustering.cluster.dbscan import dbscan
 from pyclustering.cluster.cure import cure
 from pyclustering.cluster.fcm import fcm
 from pyclustering.cluster.somsc import somsc
@@ -44,6 +46,7 @@ import copy
 import os
 from tqdm import tqdm
 import warnings
+from time import sleep
 
 
 __author__ = "Eric Cacciavillani"
@@ -61,20 +64,20 @@ class AutoCluster(AutoModeler):
 
     def __init__(self,
                  df,
-                 project_sub_dir="",
-                 project_name="Auto Clustering",
+                 dataset_sub_dir="",
+                 dataset_name="Default Dataset Name",
                  overwrite_full_path=None,
                  notebook_mode=False,
-                 pca_perc=None):
+                 pca_perc=1.00):
         """
         Args:
             df: pd.Dataframe
                 pd.Dataframe
 
-            project_sub_dir: string
+            dataset_sub_dir: string
                 Sub directory to write data.
 
-            project_name: string
+            dataset_name: string
                 Main project directory
 
             overwrite_full_path: string
@@ -85,7 +88,7 @@ class AutoCluster(AutoModeler):
         """
 
         AutoModeler.__init__(self,
-                             f'{project_sub_dir}/{project_name}',
+                             f'{dataset_name}/{dataset_sub_dir}',
                              overwrite_full_path)
 
 
@@ -93,6 +96,15 @@ class AutoCluster(AutoModeler):
             move_folder_to_eflow_garbage(self.folder_path+"_Extras",
                                          "Auto Clustering")
 
+        # if isinstance(df, pd.DataFrame):
+        #     self.__feature_names = copy.deepcopy(list(df.columns))
+        # else:
+        #     if not feature_names:
+        #         raise UnsatisfiedRequirments("If passing in a matrix like object. "
+        #                                      "You must init feature names!")
+        #     else:
+        #         self.__feature_names = copy.deepcopy(feature_names)
+
 
         # Define model
         self.__cluster_models_paths = dict()
@@ -102,7 +114,6 @@ class AutoCluster(AutoModeler):
         self.__models_suggested_clusters = dict()
 
         self.__pca = None
-        self.__pca_perc = None
 
         self.__first_scaler = None
         self.__second_scaler = None
@@ -124,7 +135,9 @@ class AutoCluster(AutoModeler):
                                           sub_dir="PCA",
                                           filename="Applied scaler results")
 
+            print(len(scaled[0]))
             pca, scaled = self.__visualize_pca_variance(scaled)
+            print(len(scaled[0]))
 
             self.__pca = pca
 
@@ -188,6 +201,7 @@ class AutoCluster(AutoModeler):
                                                                       ".pkl"):
 
                                 model_path = self.folder_path + f"Models/{cluster_dir}/{dir}/{model_name}"
+                                print(model_path)
 
                                 try:
                                     obj = load_pickle_object(model_path)
@@ -211,7 +225,7 @@ class AutoCluster(AutoModeler):
         # Save objects to directory structure
         if self.__pca:
             pipeline_path = create_dir_structure(self.folder_path,
-                                                 "Data Pipeline")
+                                                 "Data Cluster Pipeline")
 
 
             # Pickle data pipeline objects
@@ -323,7 +337,13 @@ class AutoCluster(AutoModeler):
 
 
     def visualize_hierarchical_clustering(self,
-                                          linkage_methods=None,
+                                          linkage_methods=["complete",
+                                                           "single",
+                                                           "weighted",
+                                                           "ward",
+                                                           "average",
+                                                           "centroid",
+                                                           "median"],
                                           display_print=True,
                                           display_visuals=True):
         """
@@ -645,6 +665,96 @@ class AutoCluster(AutoModeler):
 
         # self.__models_suggested_clusters["Agglomerative models"] =
 
+
+    def create_dbscan_model(self,
+                            target_clusters_amt,
+                            max_dbscan_cycle_iteration=1,
+                            max_min_sample=500):
+
+        starting_min_samples = list(range(0, max_min_sample, 50))
+        starting_min_samples.reverse()
+        starting_min_samples.remove(0)
+        max_min_sample = None
+
+        past_best_min_sample = None
+        past_best_ep = None
+
+        for min_sample in starting_min_samples:
+            best_ep, _ = self.__dbscan_find_best_ep(max_cycle=1,
+                                                    ep_incremental=.05,
+                                                    best_min_sample=min_sample,
+                                                    dbscan_cycle_iteration="Finding Max Min Sample")
+
+            if best_ep != float("-inf"):
+                best_min_sample = min_sample
+                max_min_sample = min_sample
+                break
+
+        best_ep = float("-inf")
+        highest_silhouette_avg = float("-inf")
+
+        print("Starting DBScan cycles")
+        if max_dbscan_cycle_iteration < 1:
+            print("max_dbscan_cycle_iteration must be at least 1. Converting now.")
+            max_dbscan_cycle_iteration = 1
+
+        sleep(.1)
+
+        for dbscan_cycle_iteration in range(1, max_dbscan_cycle_iteration + 1):
+            highest_silhouette_avg = 0
+            best_ep, highest_silhouette_avg = self.__dbscan_find_best_ep(max_cycle=4,
+                                                                         ep_incremental=.05,
+                                                                         best_min_sample=best_min_sample,
+                                                                         best_ep=best_ep,
+                                                                         highest_silhouette_avg=highest_silhouette_avg,
+                                                                         dbscan_cycle_iteration=dbscan_cycle_iteration)
+
+            best_min_sample = self.__dbscan_find_best_min_sample(best_ep,
+                                                                 target_clusters_amt,
+                                                                 max_min_sample,
+                                                                 dbscan_cycle_iteration)
+
+            if best_ep == past_best_ep and best_min_sample == past_best_min_sample:
+                print("General Convergence")
+                break
+
+            past_best_min_sample = best_min_sample
+            past_best_ep = best_ep
+
+            print("----------------")
+            print(best_ep)
+            print(best_min_sample)
+            print("----------------\n")
+            sleep(.1)
+
+        # create object that uses CCORE for processing
+        best_ep = round(best_ep, 10)
+        dbscan_instance = dbscan(self.__scaled, best_ep, best_min_sample, True)
+        # cluster analysis
+        dbscan_instance.process()
+        clusters = dbscan_instance.get_clusters()
+        cluster_amt = len(clusters)
+
+        model_path = create_dir_structure(self.folder_path,
+                                          f"Models/DBScan/Clusters={cluster_amt}")
+
+        # Save model and meta data
+        _ = pickle_object_to_file(dbscan_instance,
+                                  model_path,
+                                  f"DBScan_Clusters={str(cluster_amt)}")
+
+        dict_to_json_file({"ep":float(best_ep),
+                           "min_sample":int(best_min_sample)},
+                            model_path,
+                           "Parameters")
+
+
+
+        pickle_object_to_file(self.__get_centers(dbscan_instance),
+                              model_path,
+                              f"Center points")
+
+
     def evaluate_all_models(self,
                             df,
                             df_features,
@@ -1106,6 +1216,7 @@ class AutoCluster(AutoModeler):
                 # Run cluster analysis and obtain results.
                 model.process()
                 final_centers = np.array(self.__get_centers(model))
+                # print(f"final_centers {final_centers}")
                 labels = [self.__nearest_cluster(final_centers, dp)
                           for dp in self.__scaled]
 
@@ -1143,13 +1254,19 @@ class AutoCluster(AutoModeler):
             center_points = []
 
             for cluster_indexes in model.get_clusters():
-                all_dps = np.matrix([self.__scaled[i] for i in cluster_indexes])
+                # all_dps = np.matrix([self.__scaled[i] for i in cluster_indexes])
+                all_dps = np.array(
+                    [self.__scaled[i] for i in cluster_indexes])
+
+                # print(all_dps)
+                # print("----------\n")
+                # print(np.array([self.__scaled[i] for i in cluster_indexes]))
                 center_dp = all_dps.mean(0)
 
                 # Grave Yard code: Use existing point rather than generating abstract average data point
                 # np.absolute(all_dps - center_dp).sum(1).argmin()
 
-                center_points.append(np.array(center_dp.tolist()[0]))
+                center_points.append(np.array(center_dp))
 
             return center_points
 
@@ -1309,13 +1426,14 @@ class AutoCluster(AutoModeler):
 
             # Get inertia values
             if isinstance(elbow_inertias_matrix, type(None)):
-                inertias_matrix = np.matrix([inertias[i]])
-                elbow_inertias_matrix = np.matrix(inertias[i][elbow_cluster - 2:elbow_cluster + 1])
+                inertias_matrix = np.array([inertias[i]])
+                elbow_inertias_matrix = np.array(
+                    [inertias[i][elbow_cluster - 2:elbow_cluster + 1]])
             else:
                 inertias_matrix = np.vstack([inertias_matrix, inertias[i]])
-
                 elbow_inertias_matrix = np.vstack(
-                    [elbow_inertias_matrix, inertias[i][elbow_cluster - 2:elbow_cluster + 1]])
+                    [elbow_inertias_matrix, inertias[i][elbow_cluster - 2:
+                                                        elbow_cluster + 1]])
 
             elbow_models.append(k_models[i][elbow_cluster - 2:elbow_cluster + 1])
 
@@ -1370,8 +1488,6 @@ class AutoCluster(AutoModeler):
         plt.ylabel('Inertia')
         plt.xticks(ks)
 
-
-
         # Find the most agreed upon elbow section
         average_elbow_inertias = elbow_inertias_matrix.mean(0)
 
@@ -1382,11 +1498,11 @@ class AutoCluster(AutoModeler):
 
         best_elbow_index = np.array(elbow_vote).argmin()
 
-        print(inertias_matrix[best_elbow_index].tolist()[0])
+        print(inertias_matrix[best_elbow_index].tolist())
 
         # Plot most agreed upon elbow section
         plt.plot(ks,
-                 inertias_matrix[best_elbow_index].tolist()[0],
+                 inertias_matrix[best_elbow_index].tolist(),
                  '-o',
                  color='#367588')
 
@@ -1420,7 +1536,7 @@ class AutoCluster(AutoModeler):
                 print(f"Something went wrong when trying to save the model: {model_name}")
 
             plt.plot(ks[k_val - 1],
-                     inertias_matrix[best_elbow_index].tolist()[0][k_val - 1],
+                     inertias_matrix[best_elbow_index].tolist()[k_val - 1],
                      'r*')
             best_clusters.append(k_val)
 
@@ -1464,10 +1580,10 @@ class AutoCluster(AutoModeler):
 
             all_clusters += best_clusters
 
-
-        write_object_text_to_file(round(sum(all_clusters) / len(all_clusters)),
-                                  self.folder_path + "_Extras",
-                                  "Average of suggested clusters")
+        if all_clusters:
+            write_object_text_to_file(round(sum(all_clusters) / len(all_clusters)),
+                                      self.folder_path + "_Extras",
+                                      "Average of suggested clusters")
 
 
     def __get_unique_random_indexes(self,
@@ -1604,6 +1720,104 @@ class AutoCluster(AutoModeler):
 
         plt.close("all")
 
+
+    def __dbscan_find_best_ep(self,
+                              max_cycle=3,
+                              ep_incremental=.05,
+                              best_min_sample=50,
+                              best_ep=float("-inf"),
+                              highest_silhouette_avg=float("-inf"),
+                              dbscan_cycle_iteration=None):
+
+        pbar = tqdm(total=max_cycle)
+
+        for cycle_count in range(1, max_cycle + 1):
+            new_highest_found = False
+
+            if cycle_count == 1:
+                all_ep_values = np.arange(.05, 1.01, ep_incremental)
+            else:
+                all_ep_values = np.arange(best_ep - (ep_incremental),
+                                          best_ep + ep_incremental,
+                                          ep_incremental / 10)
+                ep_incremental /= 10
+
+            for ep_value in all_ep_values:
+                # create object that uses CCORE for processing
+
+                ep_value = round(ep_value, 10)
+                dbscan_instance = dbscan(self.__scaled, ep_value, best_min_sample,
+                                         True)
+                dbscan_instance.process()
+                clusters = dbscan_instance.get_clusters()
+
+                pbar.set_description(
+                    f"DBScan Cycle Iteration: {dbscan_cycle_iteration}   Epsilon Value: {ep_value}     Highest Silhouette Avg = {highest_silhouette_avg}",
+                    refresh=False)
+                pbar.update(1)
+
+                if len(clusters) <= 1:
+                    continue
+
+                labels = []
+                new_data = []
+                for i, dp_list in enumerate(clusters):
+                    for dp_index in dp_list:
+                        new_data.append(self.__scaled[dp_index])
+                    labels += [i for _ in dp_list]
+                silhouette_avg = silhouette_score(new_data, labels)
+
+                if silhouette_avg > highest_silhouette_avg:
+                    highest_silhouette_avg = silhouette_avg
+                    best_ep = ep_value
+                    new_highest_found = True
+
+            if not new_highest_found:
+                break
+
+        pbar.close()
+        return best_ep, highest_silhouette_avg
+
+    def __dbscan_find_best_min_sample(self,
+                                      best_ep,
+                                      target_clusters_amt,
+                                      max_min_sample,
+                                      dbscan_cycle_iteration):
+
+        pbar = tqdm(total=max_min_sample)
+        best_min_sample = None
+
+        cluster_amts = []
+        all_min_samples = [i for i in range(1, max_min_sample + 1)]
+        for i, min_sample in enumerate(all_min_samples):
+            pbar.set_description(
+                f"DBScan Cycle Iteration: {dbscan_cycle_iteration}   Min Sample: {min_sample}",
+                refresh=False)
+            pbar.update(1)
+
+            # create object that uses CCORE for processing
+            best_ep = round(best_ep, 10)
+            dbscan_instance = dbscan(self.__scaled, best_ep, min_sample, True)
+            # cluster analysis
+            dbscan_instance.process()
+            clusters = dbscan_instance.get_clusters()
+            cluster_amt = len(clusters)
+
+            cluster_amts.append(cluster_amt)
+
+            if cluster_amt == 2 or cluster_amt == target_clusters_amt:
+                pbar.update(max_min_sample - i - 1)
+                cluster_amts = np.asarray(cluster_amts)
+                best_min_sample = (np.abs(
+                    cluster_amts - target_clusters_amt)).argmin() + 1
+                pbar.set_description(
+                    f"DBScan Cycle: {dbscan_cycle_iteration} Best Min Sample Found: {best_min_sample}",
+                    refresh=False)
+                break
+
+        pbar.close()
+        return best_min_sample
+
     def __get_color_name(self,
                          color):
         """
